"""
setup.py - Kept for backward compatibility
Modern configuration is in pyproject.toml
"""
from setuptools import setup

# Configuration is now in pyproject.toml
# This file is kept for backward compatibility
setup()